package Library;

import VisitorLib._
import BinaryTree._
import InternalVisitors._

object BinaryTreeExample {
    def depth = new CaseTree[Internal,Int] {
      def Empty = 0
      def Fork (x : Int, l : Int,r : Int) = 1 + max(l,r)
    }
    
    def depth2 : Tree => Int = new CaseTree[External,Int] {
      def Empty                            = 0
      def Fork (x : Int, l : Rec,r : Rec)  = 1 + max(depth2(l),depth2(r))
    }
    
    def depth3 : CaseTree[External,Int] = new CaseTree[External,Int] {
      def Empty = 0
      def Fork (x : Int, l : R[TreeVisitor],r : R[TreeVisitor]) = 1 + max(l.accept(this),r.accept(this))
    } 
    
    class AddTree extends CaseTree[Internal, Unit] {
      var sumValue = 0
      
      def Empty = {}
      def Fork(x : Int, l : Unit, r : Unit) = {sumValue += x;}
    }
    
    class AddTree2 extends CaseTree[External,Unit] {
      var sumValue = 0
      
      def Empty                            = {}
      def Fork(x : Int, l : Rec, r : Rec)  = 
        {this(l);this(r); sumValue += x;}
    }
    
    def test : Int = {
      val addTree  = new AddTree();
      val tree1    = Fork(3,Empty,Empty);
      val tree2    = Fork(4,tree1,tree1);
      addTree(tree2);
      
      return addTree.sumValue;
    }
     
    def addTree = new CaseTree[Internal,Int] {
      def Empty                           = 0
      def Fork (x : Int, l : Rec,r : Rec) = x + l + r
    }
    
    def addTree2 = new CaseTree[External,Int] {
      def Empty                            = 0
      def Fork (x : Int, l : Rec,r : Rec)  = x + addTree(l) + addTree(r)
    }
    
    def foldTree[a] (k : a) (f : Int => a => a => a) : CaseTree[External,a] = new CaseTree[External,a] {
        def Empty                             = k
        def Fork (x : Int, l : Rec, r : Rec)  = f (x) (foldTree(k)(f)(l)) (foldTree(k)(f)(r))
      }
    
    def buildTree (k : Tree) (f : Int => Tree => Tree => Tree)= new CaseTree[Internal,Tree] {
      def Empty = k
      def Fork (x : Int, l : Rec, r : Rec) = f (x) (l) (r)
    }
     
    def depth4 = foldTree (0) (x => l => r => 1 + max(l,r))
    
    def addTree3 (t : Tree) = foldTree (0) (x => l => r => x + l + r) (t)
      
    def max (x : Int, y : Int) = if (x > y) x else y
  
    def isEmpty : Tree => boolean = new CaseTree[External,boolean] {
      def Empty = true
      def Fork (x : Int, l : Rec, r : Rec) = false
    }
    
    def equal (t : Tree) : Tree => boolean = new CaseTree[External,Boolean] {
      def Empty = isEmpty(t)
      def Fork (x : Int, l1 : Rec, r1 : Rec) = t.accept[External,Boolean](new CaseTree[External,boolean] {
      	def Empty = false
        def Fork (y : Int, l2 : Rec, r2 : Rec) =
          x == y && equal(l1)(l2) && equal(r1)(r2)
      })
    }
    
    def tree1 = Fork(3,Empty,Empty)
    
    def tree2 = Fork(4,tree1,tree1)
}
