package Library;

import VisitorLib._
import ParamorphicVisitors._
import InternalVisitors._

/* Visitor for natural numbers */

object Naturals {
    /*
    data Nat {
       constructor Zero {
          val intValue = 0
       }
       
       constructor Succ (n : Nat) {
          val intValue = 1 + n.intValue
       }
       
       val intValue : int 
       
       override def toString() : String = this.accept(new CaseNat[Internal,String] {
         def Zero               = "Zero"
         def Succ (n : String)  = "Succ(" + n + ")"
       })
      
       override def equals(x : Any) : boolean = 
         x match {
           case m:Nat => intValue.equals(m.intValue)
           case _     => false
         }
      
       override def hashCode() = intValue.hashCode()
    }
    */
    
    trait NatVisitor extends Visitor {
      type Rec = R[NatVisitor]
        
      def Zero : X
      def Succ (n : Rec) : X
    }
    
    trait Nat extends Composite[NatVisitor] {
      val intValue : Int
      
      override def toString() : String = this.accept(new CaseNat[Internal,String] {
        def Zero               = "Zero"
        def Succ (n : String)  = "Succ(" + n + ")"
      })
      
      override def equals(x : Any) : Boolean = 
        x match {
          case m:Nat => intValue.equals(m.intValue)
          case _     => false
        }
      
      override def hashCode() = intValue.hashCode()
    }
    
    case object Zero extends Nat {
      def accept[s <: Strategy,x] (vis : VisFunc[NatVisitor,s,x]) (implicit decompose : Decompose[s]) : x =
        vis.Zero
        
      val intValue = 0
    }
    
    case class Succ(n : Nat) extends Nat {
      def accept[s <: Strategy,x] (vis : VisFunc[NatVisitor,s,x]) (implicit decompose : Decompose[s]) : x =
        vis.Succ(decompose.dec(vis,n))
        
      val intValue = 1 + n.intValue
    }
    
    abstract class CaseNat[s <: Strategy,x](implicit dec : Decompose[s]) extends DCaseNat[s,x](Basic)(dec)   
   
    abstract class DCaseNat[s <: Strategy, x](disp : Dispatcher[NatVisitor,s,x])(implicit dec : Decompose[s]) 
    extends Case[NatVisitor,s,x](disp)(dec) with NatVisitor
    
    // Some Examples
    def toInt = new CaseNat[Internal,Int] { 
      def Zero            = 0
      def Succ (n : Int)  = 1 + n
    }
    
    def toInt3 : CaseNat[External,int] = new CaseNat[External,Int] {
      def Zero                      = 0
      def Succ (n : R[NatVisitor])  = 1 + toInt3(n)
    }
    
    def toInt2(disp : Dispatcher[NatVisitor,Internal,Int]) = new DCaseNat[Internal,Int](disp) {
      def Zero            = 0 
      def Succ (n : Int)  = 1 + n
    }
    
    def toInt4 = toInt2(Basic)
    def toInt5 = toInt2(BeenHere(Basic))
    
    def fib1 : CaseNat[External,Int] = new CaseNat[External,int] {
      def Zero = 0
      def Succ (n : R[NatVisitor]) = n.accept(new CaseNat[External,int] {
        def Zero           = 1
        def Succ(m : R[NatVisitor])  = fib1(n) + fib1(m)
      })
    }
    
    
    def fib2(d : Dispatcher[NatVisitor,External,int]) : DCaseNat[External,int] = new DCaseNat[External,int](d) {
      def Zero = 0
      def Succ (n : Rec) = n.accept(new CaseNat[External,int] {
        def Zero           = 1
        def Succ(m : Rec)  = fib2(d)(n) + fib2(d)(m)
      })
    }
    
    def toNat (x : Int) : Nat = if (x == 0) Zero else Succ(toNat(x-1))
    
    def nfib = fib2(Basic)
    def mfib = fib2(Memo(Basic))
    def tfib = fib2(Trace(Memo(Basic)))
    def mtfib = fib2(Memo(Trace(Basic)))
    def ttfib = fib2(Trace(Basic))
    
    def test = Succ (Succ (Succ (Zero)))
    
    def seven = Succ(Succ(Succ(Succ(test))))
}
