package Library;

import scala.collection.mutable._
import scala.collection.mutable.HashMap._

/* The Visitor Library */

object VisitorLib {
  trait Strategy {
      type V <: Visitor
      type X
      type Y // Computed type
  }   
  
  trait Decompose[s <: Strategy] {
      def dec[v <: Visitor, x] (vis : VisFunc[v,s,x], comp : Composite[v]) : (s {type V=v; type X=x})#Y
  }
  
  implicit def external : Decompose[External] = new Decompose[External] {
    def dec[v <: Visitor, x] (vis : VisFunc[v,External,x], comp : Composite[v]) = comp 
  }
  
  trait Visitor {
      type X
      type S <: Strategy
      type R[v <: Visitor] = (S {type X = Visitor.this.X; type V=v})#Y
  }
  
  trait Composite[v <: Visitor] {
    def accept[s <: Strategy,x] (vis : VisFunc[v,s,x]) (implicit decompose : Decompose[s]) : x
  }

  trait External extends Strategy {
      type Y = Composite[V]
  }
    
  type VisFunc[v <: Visitor,s <: Strategy,x] = Function1[Composite[v],x] with v {type S=s; type X=x}
  
  abstract class Case[v <: Visitor,s <: Strategy,x](d : Dispatcher[v,s,x])(implicit dec : Decompose[s]) 
  extends Function1[Composite[v],x] {
    self : Case[v,s,x] with v {type S=s; type X=x} =>
    type X=x
    type S=s
    
    def dispatcher = d
    def decompose  = dec
    
    def apply(c : Composite[v]) : x = dispatcher.dispatch(this,decompose).apply(c)
  }
  
  // Dispatchers
  trait Dispatcher[v <: Visitor, s <: Strategy, x] { // avoid the parametrization?
    def dispatch(vis : VisFunc[v,s,x], dec : Decompose[s]) : Function1[Composite[v],x]
  }
  
  implicit def Basic[v <: Visitor, s <: Strategy, x] = new Dispatcher[v,s,x] {
    def dispatch(vis : VisFunc[v,s,x], dec : Decompose[s]) : Function1[Composite[v],x] = 
      c => c.accept[s,x](vis)(dec)
  }
  
  def BeenHere[v <: Visitor, s <: Strategy, x](d : Dispatcher[v,s,x]) = new Advice[v,s,x](d) {
    override def before(comp : Composite[v]) : Unit = System.out.println("Been Here!")
  }
  
  abstract class Advice[d <: Visitor, s <: Strategy, x](dis : Dispatcher[d,s,x]) extends Dispatcher[d,s,x] {
    def before(comp : Composite[d]) : Unit = {}
    def after(comp  : Composite[d], res : x) : Unit = {}

    def dispatch(vis : VisFunc[d,s,x], dec : Decompose[s]) : Function1[Composite[d],x] = 
   	  c => {before(c); val res = dis.dispatch(vis,dec)(c);after(c,res);res}
  }
  
  def Trace[v <: Visitor, s <: Strategy, x](dis : Dispatcher[v,s,x]) = new Advice[v,s,x](dis) {
    override def before(comp : Composite[v]) : Unit = {
      System.out.println("Calling function with argument: \t" + comp);
    }
    override def after(comp : Composite[v], res : x) : Unit = {
      System.out.println(res + "\t was return from the call with argument: \t" + comp);
    }
  }
  
  def Memo[v <: Visitor, s <: Strategy,x](dis : Dispatcher[v,s,x]) = new Dispatcher[v,s,x] {
    val cache : HashMap[Composite[v],x] = new HashMap[Composite[v],x]()
    
    def dispatch(vis : VisFunc[v,s,x], dec : Decompose[s]) : Function1[Composite[v],x] = c => {
      cache.get(c) match {
         case Some(x) => x
         case None    => {val res = dis.dispatch(vis,dec)(c); cache.put(c,res); res} 
       }
    }
  }
}