package both

import Library.VisitorLib._
import Library.InternalVisitors._

object Components {
  
  // Base component
  type Expr[V <: Visitor] = Composite[V]
  
  // Numeric components
  trait vnum extends Visitor {
    def num(x : Int) : X
  }
  
  case class Num[V <: vnum](x : Int) extends Expr[V] {
    def accept[s <: Strategy,x] (vis : VisFunc[V,s,x]) (implicit decompose : Decompose[s]) : x =
        vis.num(x)
  }
 
  // Addition components
  trait vadd[V <: Visitor] extends Visitor {
    def add(e1 : R[V], e2 : R[V]) : X
  }
  
  case class Add[V <: vadd[V]](e1 : Expr[V], e2 : Expr[V]) extends Expr[V] {
    def accept[s <: Strategy,x] (vis : VisFunc[V,s,x]) (implicit decompose : Decompose[s]) : x =
        vis.add(decompose.dec(vis,e1),decompose.dec(vis,e2))
  }
  
  // Negation components
  trait vneg[V <: Visitor] extends Visitor {
    def neg(e : R[V]) : X
  }
    
  case class Neg[V <: vneg[V]](e : Expr[V]) extends Expr[V] {
    def accept[s <: Strategy,x] (vis : VisFunc[V,s,x]) (implicit decompose : Decompose[s]) : x =
        vis.neg(decompose.dec(vis,e))
  }
  
  // some auxiliary visitor stuf
  abstract class CaseTree[v <: Visitor,s <: Strategy,x](implicit dec : Decompose[s]) extends DCaseExpr[v,s,x](Basic)(dec) {
    self : v {type X=x; type S=s} =>
  }
    
  abstract class DCaseExpr[v <: Visitor,s <: Strategy,x](disp : Dispatcher[v,s,x])(implicit dec : Decompose[s]) 
      extends Case[v,s,x](disp)(dec) {
        self : v {type X=x; type S=s} =>
  }
  
  // Evaluation component
  trait BaseEval[V <: Visitor] extends CaseTree[V,Internal,Int] with vnum with vadd[V] with vneg[V] {
    self : V {type X=Int;type S=Internal} =>
    
    def num(x : Int)             = x
    def add(e1 : Int, e2 : Int)  = e1 + e2 
    def neg(e : Int)             = -e
  }
  
  // Pretty printing component
  trait BasePrint[V <: Visitor] extends CaseTree[V,External,String] with vnum with vadd[V] with vneg[V] {
    self : V {type X=String; type S=External} =>
    
    def num(x : Int)               = x.toString()
    def add(e1 : R[V], e2 : R[V])  = "(" + this(e1) + " + " + this(e2) + ")"
    def neg(e : R[V])              = "-" + this(e)
  }
}