package external

import Components._

object ExtendedComponents {
  // Specification of what we want
  
  // eq : Expr -> Expr -> Expr
  // eq (Num n1)    (Num n2)     = n1.equals(n2)
  // eq (Add e1 e2) (Add e3 e4)  = eq(e1)(e3) && eq(e2)(e4)
  // eq (Neg e1)    (Neg e2)     = eq(e1)(e2)
  // eq _           _            = false
  
  // Prototypes components for equality
  trait BaseHandleDefault[V[-_,_],A] {
    self : V[Expr[V],A => Boolean] =>
    
    // recursive call reference 
    def eqVis  : V[Expr[V],Expr[V] => Boolean]
    
    // default value
    val default = (_:A) => false
  }
  
  // Numeric Components for equality
  trait NumHandleDefault[V[-_,_],A] extends BaseHandleDefault[V,A] with vnum[Expr[V],A => Boolean] {
    self : V[Expr[V],A => Boolean] =>
    
    def num(n2 : Int) = default
  }
  
  trait NumEquals[V[-_,_]] extends vnum[Expr[V],Expr[V] => Boolean] {
    self : V[Expr[V],Expr[V] => Boolean] =>
    
    def eqNum : V[Expr[V],Int => Boolean]
    
    def num(n : Int) = e => e.accept(eqNum)(n)
  }
  
  trait HandleNum[V[-R,A]] extends NumHandleDefault[V,Int] {
    self : V[Expr[V],Int => Boolean] =>
                    
    override def num(n2 : Int) = n1 => n1.equals(n2)
  }
  
  // Addition components for equality
  type Tuple[V[-_,_]] = (Expr[V],Expr[V])
  
  trait AddHandleDefault[V[-_,_],A] extends BaseHandleDefault[V,A] with vadd[Expr[V],A => Boolean] {
    self : V[Expr[V],A => Boolean] =>
    
    def add(e1 : Expr[V], e2 : Expr[V])  = default
  }
  
  trait AddEquals[V[-_,_]] extends vadd[Expr[V],Expr[V] => Boolean] {
    self : V[Expr[V],Expr[V] => Boolean] =>
    
    def eqAdd : V[Expr[V],Tuple[V] => Boolean]
    def add(e1 : Expr[V], e2 : Expr[V])  = e  => e.accept(eqAdd)(e1,e2)
  }
  
  trait HandleAdd[V[-_,_]] extends AddHandleDefault[V,Tuple[V]] {
    self : V[Expr[V],Tuple[V] => Boolean] =>  
    
    override def add(e3 : Expr[V], e4 : Expr[V])  = t => 
      t._1.accept(eqVis)(e3) && t._2.accept(eqVis)(e4)
  }
}