package external

object Features {
  trait Datatype {
    protected val C=Components
    
    protected type ExprVisitor[-R,A] 
    type Expr = C.Expr[ExprVisitor]           
  }
  
  // Numeric Feature
  trait Numeric extends Datatype {
    type ExprVisitor[-R,A] <: C.vnum[R,A] // constraints of the feature
    
    def Num(x : Int) = C.Num[ExprVisitor](x)
  } 
  
  // Addition Feature
  trait Addition extends Datatype {
    type ExprVisitor[-R,A] <: C.vadd[R,A]
    
    def Add(e1 : Expr, e2 : Expr)  = C.Add[ExprVisitor](e1,e2)
  }
  
  // Negation Feature
  trait Negation extends Datatype {
    type ExprVisitor[-R,A] <: C.vneg[R,A]
    
    def Neg(e : Expr)  = C.Neg[ExprVisitor](e)
  }
  
  // Eval Feature
  trait Eval extends Datatype {
    protected type BaseEval     = C.BaseEval[ExprVisitor]
    protected type EvalVisitor  = BaseEval with ExprVisitor[Expr,Int] 
      
    protected val evalVisitor : EvalVisitor // abstract
    
    def eval(e : Expr) : Int = e.accept[Int](evalVisitor)
  }
  
  // Print Feature
  trait Print extends Datatype {
    type PrintVisitor = C.BasePrint[ExprVisitor] with ExprVisitor[Expr,String]
    
    protected val printVisitor : PrintVisitor // abstract
    
    def prettyPrint(e : Expr) : String = e.accept[String](printVisitor)
  }
  
  // Narrow Feature
  trait Narrow extends Datatype {
    type TExpr = C.Expr[TExprVisitor]
    protected type TExprVisitor[-R,A]
    
    type NarrowVisitor = ExprVisitor[Expr,TExpr] 
    
    protected val narrowVisitor : NarrowVisitor // abstract
    
    def narrow(e: Expr) : TExpr = e.accept[TExpr](narrowVisitor)
  }
  
  // Just adds constraints
  trait NumNarrow extends Numeric with Narrow {
    protected type TExprVisitor[-R,A] <: C.vnum[R,A] 
    protected type NumNarrow = C.NumNarrow[ExprVisitor,TExprVisitor]
  }
  
  trait AddNarrow extends Addition with Narrow {
    protected type TExprVisitor[-R,A] <: C.vadd[R,A]
    
    protected type AddNarrow = C.AddNarrow[ExprVisitor,TExprVisitor]
  }
  
  trait NegNarrow extends Negation with Narrow {
    protected type TExprVisitor[-R,A] <: C.vnum[R,A] with C.vminus[R,A]
    
    protected type NMNarrow = C.NMNarrow[ExprVisitor,TExprVisitor]
  }
  
  // Equality Feature
  trait Eq extends Datatype {
    protected val EC=ExtendedComponents
    
    type EqualsVisitor = ExprVisitor[Expr,Expr => Boolean]
    
    trait EqualsDefault[A] extends EC.BaseHandleDefault[ExprVisitor,A] {
      self : ExprVisitor[Expr,A => Boolean] => 
    }
    
    trait Equals {self : EqualsVisitor => }
    
    protected val equalsVisitor : EqualsVisitor // abstract
    
    def equalsExpr(e1 : Expr, e2 : Expr) : Boolean = e1.accept[Expr => Boolean](equalsVisitor)(e2)
  }
  
  trait NumericEquality extends Numeric with Eq {
    type EqualsNum = EC.HandleNum[ExprVisitor] with ExprVisitor[Expr,Int => Boolean]
    
    trait EqualsDefault[A] extends EC.NumHandleDefault[ExprVisitor,A] with super.EqualsDefault[A] {
      self : ExprVisitor[Expr,A => Boolean] =>
    }
    
    trait Equals extends EC.NumEquals[ExprVisitor] with super.Equals { // requires deep mixin composition?
      self : ExprVisitor[Expr, Expr => Boolean] =>
    
      def eqNum = handleNum
    }
    
    trait HandleNum extends EC.HandleNum[ExprVisitor] {
      self : ExprVisitor[Expr, Int => Boolean] =>
      
      def eqVis = equalsVisitor
    }
    
    protected val handleNum : EqualsNum // abstract
  }
  
  trait AdditionEquality extends Addition with Eq {
    type EqualsAdd     = EC.HandleAdd[ExprVisitor] with ExprVisitor[Expr,EC.Tuple[ExprVisitor] => Boolean]
    
    trait EqualsDefault[A] extends EC.AddHandleDefault[ExprVisitor,A] with super.EqualsDefault[A] {
      self : ExprVisitor[Expr,A => Boolean] =>
    }
    
    trait Equals extends EC.AddEquals[ExprVisitor] with super.Equals { // requires deep mixin composition?
      self : ExprVisitor[Expr, Expr => Boolean] =>
      
      def eqAdd = handleAdd
    }
    
    trait HandleAdd extends EC.HandleAdd[ExprVisitor] {
      self : ExprVisitor[Expr, EC.Tuple[ExprVisitor] => Boolean] =>
      
      def eqVis = equalsVisitor
    }
    
    protected val handleAdd : EqualsAdd // abstract
  }
  
  // Too much boilerplate!
  trait NumAddEquality extends NumericEquality with AdditionEquality {
    type ExprVisitor[-R,A] <: C.vnum[R,A] with C.vadd[R,A] // I need to explicitly provide this
    
    trait EqualsDefault[A] extends super[NumericEquality].EqualsDefault[A] with super[AdditionEquality].EqualsDefault[A]  {
      self : ExprVisitor[Expr,A => Boolean] =>
    }
    
    trait Equals extends super[NumericEquality].Equals with super[AdditionEquality].Equals {
      self : ExprVisitor[Expr, Expr => Boolean] =>
    }
    
    trait HandleAdd extends EqualsDefault[EC.Tuple[ExprVisitor]] with super.HandleAdd {
      self : ExprVisitor[Expr, EC.Tuple[ExprVisitor] => Boolean] =>
    }
    
    trait HandleNum extends EqualsDefault[Int] with super.HandleNum {
      self : ExprVisitor[Expr, Int => Boolean] =>
    }
  }
  
  // Equality Feature
  // Maybe a good example of something evolving covariantly
  trait Equality extends Datatype {// There is some entanglement here
    protected val EC=ExtendedComponents
    
    type EqualsVisitor = /*EC.AddEquals[ExprVisitor] with EC.NumEquals[ExprVisitor] with*/ ExprVisitor[Expr,Expr => Boolean]
    type EqualsNum     = EC.HandleNum[ExprVisitor] with ExprVisitor[Expr,Int => Boolean]
    type EqualsAdd     = EC.HandleAdd[ExprVisitor] with ExprVisitor[Expr,EC.Tuple[ExprVisitor] => Boolean]
   
    trait EqualsDefault[A] extends EC.AddHandleDefault[ExprVisitor,A] with EC.NumHandleDefault[ExprVisitor,A] {
      self : ExprVisitor[Expr,A => Boolean] =>
    }
    
    trait Equals extends EC.AddEquals[ExprVisitor] with EC.NumEquals[ExprVisitor] { // requires deep mixin composition?
      self : ExprVisitor[Expr, Expr => Boolean] =>
      
      // this makes it dependent on Num and Add
      def eqNum = handleNum
      def eqAdd = handleAdd
    }
    
    // These can be defined elsewhere
    trait HandleNum extends EqualsDefault[Int] with EC.HandleNum[ExprVisitor] {
      self : ExprVisitor[Expr, Int => Boolean] =>
      
      def eqVis = equalsVisitor
    }
    
    trait HandleAdd extends EqualsDefault[EC.Tuple[ExprVisitor]] with EC.HandleAdd[ExprVisitor] {
      self : ExprVisitor[Expr, EC.Tuple[ExprVisitor] => Boolean] =>
      
      def eqVis = equalsVisitor
    }
    
    protected val equalsVisitor : EqualsVisitor // abstract
    protected val handleNum : EqualsNum // abstract
    protected val handleAdd : EqualsAdd // abstract
    
    def equalsExpr(e1 : Expr, e2 : Expr) : Boolean = e1.accept[Expr => Boolean](equalsVisitor)(e2)
  }
}
