package internal

object Components {
  // The expression component
  trait Expr[-V[_]] { 
    def accept[a](vis : V[a]) : a
  }
  
  // Components for numeric expressions
  trait vnum[A] {
    def num(x : Int) : A
  }
  
  case class Num[-V[X] <: vnum[X]](x : Int) extends Expr[V] {
    def accept[a](vis : V[a]) : a = vis.num(x)
  }
  
  // Components for addition expressions
  trait vadd[A] {
    def add(e1 : A, e2 : A) : A
  }
  
  case class Add[-V[X] <: vadd[X]](e1 : Expr[V], e2 : Expr[V]) extends Expr[V] {
    def accept[a](vis : V[a]) : a = vis.add(e1.accept(vis),e2.accept(vis)) 
  }
  
  // Evaluation component
  trait BaseEval extends vnum[Int] with vadd[Int] { 
    def num(x : Int)             = x
    def add(e1 : Int, e2 : Int)  = e1 + e2 
  }
  
  // Print Visitor
  trait BasePrint extends vnum[String] with vadd[String] {
    def num(x : Int)                   = x.toString()
    def add(s1 : String, s2 : String)  = "(" + s1 + " + " + s2 + ")"  
  }
}