package internal

object Features { 
  trait Datatype {
    protected val C=Components
    
    type ExprVisitor[A] 
    type Expr = C.Expr[ExprVisitor]           
  }
  
  // Numeric Feature
  trait Numeric extends Datatype {
    type ExprVisitor[A] <: C.vnum[A] // constraints of the feature
    
    def Num(x : Int) = C.Num[ExprVisitor](x)
  } 
  
  // Addition Feature
  trait Addition extends Datatype {
    type ExprVisitor[A] <: C.vadd[A] // constraints of the feature
    
    def Add(e1 : Expr, e2 : Expr)  = C.Add[ExprVisitor](e1,e2)
  }
  
  // Eval Feature
  trait Eval extends Datatype {
    type EvalVisitor = C.BaseEval with ExprVisitor[Int]
    
    protected val evalVisitor : EvalVisitor // abstract
    
    def eval(e : Expr) : Int = e.accept[Int](evalVisitor)
  }
  
  // Print Feature
  trait Print extends Datatype {
    type PrintVisitor = C.BasePrint with ExprVisitor[String]
    
    protected val printVisitor : PrintVisitor // abstract
    
    def prettyPrint(e : Expr) : String = e.accept[String](printVisitor)
  }
}
